/*
	
	DragView.cpp
	
	Copyright 1995 Be Incorporated, All Rights Reserved.
	
*/

#define EBITMAP_ARCHIVE_DEMO

#ifndef Drag_VIEW_H
#include "DragView.h"
#endif

#ifndef Drag_WINDOW_H
#include "DragWindow.h"
#endif

#ifndef EBITMAP_LIB_H
#include "ebitmap_lib.h"
#endif

#ifndef EBITMAP_H
#include "EBitmap.h"
#endif

#ifdef EBITMAP_ARCHIVE_DEMO

#pragma export on

class EBitmap : public BArchivable {
	BMessage msg;
	EBitmap ();
	
public:
	EBitmap (BMessage *from);
	virtual ~EBitmap ();
	
	static BArchivable *Instantiate (BMessage *from);
	
	BBitmap *MakeABitmap ();
};

#pragma export reset

#endif

DragView::DragView(BRect rect, char *name, BBitmap *bitmap)
	: BStringView(rect, name, "hello!", B_FOLLOW_ALL, B_WILL_DRAW), bm(bitmap) {
}

void DragView::AttachedToWindow () {
	BFont font (be_bold_font);
	
	font.SetSize (18);
	SetFont (&font);
}

void DragView::Draw(BRect r) {
	
	if (bm)
		DrawBitmap(bm);
	else {
		r = Bounds ();
		r.InsetBy (1, 1);
		StrokeRect (r);
		MovePenTo (10, 40);
		DrawString ("Drop EBitmap here!");
	}
}

BPoint newLocation (20, 80);

void DragView::MessageReceived(BMessage *msg) {
	BBitmap *bitmap;
	
	if (msg->what == E_BITMAP) {
#ifdef EBITMAP_ARCHIVE_DEMO
		EBitmap *eBmp = dynamic_cast<EBitmap *>(instantiate_object (msg));
		if (eBmp != NULL) {
			bitmap = eBmp->MakeABitmap ();
		}
#else
		bitmap = ebitmap_to_bbitmap (msg);
#endif
		if (bitmap != NULL) {
			BRect r(bitmap->Bounds ());
			r.OffsetTo (newLocation);
			newLocation += BPoint (20, 20);
			if (newLocation.y > 200) {
				newLocation.Set (20, 80);
			}
			DragWindow *wind = new DragWindow (r, bitmap);
			wind->Show ();
		}
	}
	else {
		BView::MessageReceived (msg);
	}
}

#ifdef EBITMAP_ARCHIVE_DEMO
EBitmap::EBitmap (BMessage *from) {

	// make our message empty and copy over the other stuff
	msg.MakeEmpty ();
	msg = *from;
}

EBitmap::~EBitmap () {
}

BArchivable *EBitmap::Instantiate (BMessage *from) {
	
	if (!validate_instantiation (from, "EBitmap")) 
		return NULL; 
	return new EBitmap (from);
}

BBitmap *EBitmap::MakeABitmap () {

	return ebitmap_to_bbitmap (&msg);
}
#endif
