//#define DEMO_5

/*
	
	HelloWorld.cpp
	
	Copyright 1995 Be Incorporated, All Rights Reserved.
	
*/

#ifndef HELLO_WINDOW_H
#include "HelloWindow.h"
#endif
#ifndef HELLO_VIEW_H
#include "HelloView.h"
#endif
#ifndef HELLO_WORLD_H
#include "HelloWorld.h"
#endif

#ifndef _DATATYPES_H_
#include <Datatypes.h>
#endif
#ifndef EBITMAP_H
#include "EBitmap.h"
#endif

#define debug_printf(s,n) { char buf[256]; sprintf(buf, s, n); debugger(buf); }

main() {	
	HelloApplication *myApplication;

	myApplication = new HelloApplication ();
	myApplication->Run();
	
	delete(myApplication);
	return(0);
}

HelloApplication::HelloApplication() 
	: BApplication("application/x-vnd.GPC-HelloWorld"),
	datatypesInitialized(false),
	imageList( new BList (1)) {
}

HelloApplication::~HelloApplication() {

	delete imageList;
}


void HelloApplication::ReadyToRun () {
	status_t err;
	
	if (imageList->CountItems () > 0) {
		if (DATAVersion != NULL) {
			long curVersion, minVersion;
			DATAVersion (curVersion, minVersion);
			if ((minVersion <= DATA_CURRENT_VERSION) && (curVersion >= DATA_MIN_VERSION)) {
				err = DATAInit ("application/x-vnd.GPC-dt");
				if (err == B_OK) {
					datatypesInitialized = true;
					imageList->DoForEach (HelloApplication::CreateImage);
					imageList->DoForEach (HelloApplication::ItemDeleter);
					delete imageList;
					imageList = NULL;
					return;
				}
			}
		}
	}
	// if anything bad happens QUIT!
	be_app->PostMessage (B_QUIT_REQUESTED);
}

bool HelloApplication::QuitRequested () {

	if (datatypesInitialized) DATAShutdown ();
	
	return BApplication::QuitRequested ();
}

void HelloApplication::RefsReceived (BMessage *msg) {
	entry_ref *the_ref;
	uint32 type;
	int32 count;
	
	msg->GetInfo ("refs", &type, &count);
	if (type != B_REF_TYPE)
		return;
	for (int i=0;i<count;i++) {
		the_ref = new entry_ref ();
		if (msg->FindRef ("refs", i, the_ref) == B_OK) {
			if (imageList != NULL) {
				// if imageList isn't NULL we're starting up otherwise
				// someone dropped a file on us while we are running
				// see Constructor and ReadyToRun
				imageList->AddItem (the_ref);
			}
			else {
				CreateImage (the_ref);
			}
		}
		else {
			// Opps, we didn't need this one
			delete the_ref;
		}
	}
}

bool HelloApplication::ItemDeleter (void *v) {
	entry_ref *ref = static_cast<entry_ref *>(v);
	
	if (ref) {
		delete ref;
	}
	return false;
}

#if !defined(DEMO_5)
bool HelloApplication::CreateImage (void *v) {
	HelloWindow		*aWindow;
	HelloView		*aView;
	BRect			aRect;
	status_t err;
	char *outName;
	char *info;
	int32 outVersion;
	DATAInfo *outInfo = NULL;
	int32 outNumInfo;
	int32 whichOne = -1;
	entry_ref *the_ref = (entry_ref *)(v);
	
	if (the_ref != NULL) {
		BFile inSource (the_ref, O_RDONLY);
		if (inSource.InitCheck () == B_OK) {
			BMallocIO outSource;
			err = DATAGetHandlers (inSource, NULL, outInfo, outNumInfo, 'TIFF');
			if (err == B_OK) {
				for (int i=0;i<outNumInfo;i++) {
					err = DATAGetHandlerInfo(outInfo[i].formatHandler, outName, info, outVersion);
					if (err == B_OK) {
						if (strcmp (outName, "TIFF EBitmap Handler") == 0) {
							whichOne = i;
							break;
						}
					}
				}
				if (whichOne != -1) {
					err = DATATranslate (inSource, &outInfo[whichOne], NULL, outSource, E_BITMAP, 0, "image/tiff");
				}
				else {
					err = DATATranslate (inSource, NULL, NULL, outSource, E_BITMAP, 0, "image/tiff");
				}
				// we can finally delete this
				delete [] outInfo;
			}
			else {
				err = DATATranslate (inSource, NULL, NULL, outSource, E_BITMAP, 0, "image/tiff");
			}
			if (err == B_OK) {
				// translation was OK
				BMessage eBmp;
				int32 width;
				int32 height;
				int32 photometric;
				int32 rowBytes;
				int32 storage;
				
				outSource.Seek (0, SEEK_SET); // This is important!
				if ((err = eBmp.Unflatten (&outSource)) == B_OK) {
					if (eBmp.FindInt32 ("width", &width) == B_OK) { // required
						if (eBmp.FindInt32 ("height", &height) == B_OK) { // required
							if ((eBmp.FindInt32 ("photometric", &photometric) == B_OK) && (photometric == B_RGB_32_BIT)) { // required
								if (eBmp.FindInt32 ("storage", &storage) == B_OK) { // required
									if (eBmp.FindInt32 ("row bytes", &rowBytes) != B_OK) { // probably a good idea
										rowBytes = width*sizeof (rgb_color);
									}
									BBitmap *bm = new BBitmap (BRect (0, 0, width-1, height-1), static_cast<color_space>(photometric));
									// fill in the rest of the BBitmap here
									char *buf = static_cast<char *>(bm->Bits ());
									if (buf != NULL) {
										switch (storage) {
											case E_STORAGE_FOLLOWS:
												ssize_t bytesRead = outSource.Read (buf, rowBytes*height);
												break;
											case E_STORAGE_POINTER:
												void *data;
												err = eBmp.FindPointer ("pointer", &data);
												if (err == B_OK) {
													memcpy (buf, data, rowBytes*height);
													free (data);
												}
												break;
											case E_STORAGE_AREA_ID:
												area_id area;
												area_info info;
												err = eBmp.FindInt32 ("area_id", &area);
												if (err == B_OK) {
													err = get_area_info (area, &info);
													if (err == B_OK) {
														memcpy (buf, info.address, rowBytes*height);
														// we're done with this
														delete_area (area);
													}
												}
												break;
											case E_STORAGE_SELF:
												err = eBmp.FindData ("bits", B_RAW_TYPE, 0, &data, &bytesRead);
												if (err == B_OK) {
													memcpy (buf, data, rowBytes*height);
												}
												break;
										}
									}
//									bm->SetBits (buf, bytesRead, 0, B_RGB_32_BIT);
									aRect = bm->Bounds ();
									aRect.OffsetBy (100, 80);
									aWindow = new HelloWindow(aRect);
								
									// set up a rectangle and instantiate a new view
									// view rect should be same size as window rect but with left top at (0, 0)
									aRect = bm->Bounds ();
									aView = new HelloView(aRect, "HelloView", bm);

									// add view to window
									aWindow->AddChild(aView);
								
									// make window visible
									aWindow->Show();
									
									return false;
								}
							}
						}
					}
				}
				else {
					char buf[256];
					sprintf (buf, "Unflatten didn't work: %x", err);
					debugger (buf);
				}
			}
			else {
				char buf[256];
				sprintf (buf, "Translation didn't happen: %x", err);
				debugger (buf);
			}
		}
	}
	// something went wrong
	be_app->PostMessage (B_QUIT_REQUESTED);
	return false;
}
#else /* DEMO_5 */

// don't forget to change the PEF setting in the prefernces panel
#pragma export on
class EBitmap : public BArchivable {
	BMessage msg;
	EBitmap ();
	
public:
	EBitmap (BMessage *from);
	virtual ~EBitmap ();
	
	static BArchivable *Instantiate (BMessage *from);
	
	BBitmap *MakeABitmap ();
};
#pragma export reset

EBitmap::EBitmap (BMessage *from) {

	// make our message empty and copy over the other stuff
	msg.MakeEmpty ();
	msg = *from;
}

EBitmap::~EBitmap () {
}

BArchivable *EBitmap::Instantiate (BMessage *from) {
	
	if (!validate_instantiation (from, "EBitmap")) 
		return NULL; 
	return new EBitmap (from);
}

BBitmap *EBitmap::MakeABitmap () {
	int32 width;
	int32 height;
	int32 photometric;
	int32 rowBytes;
	int32 storage;
	BBitmap *bm = NULL;
	status_t err;
	if (msg.FindInt32 ("width", &width) == B_OK) { // required
		if (msg.FindInt32 ("height", &height) == B_OK) { // required
			if ((msg.FindInt32 ("photometric", &photometric) == B_OK) && (photometric == B_RGB_32_BIT)) { // required
				if ((msg.FindInt32 ("storage", &storage) == B_OK) && (storage == E_STORAGE_SELF)) { // required for DEMO_5 we know that this is true
					if (msg.FindInt32 ("row bytes", &rowBytes) != B_OK) { // probably a good idea
						rowBytes = width*sizeof (rgb_color);
					}
					bm = new BBitmap (BRect (0, 0, width-1, height-1), B_RGB_32_BIT);
					// fill in the rest of the BBitmap here
					if (bm != NULL) {
						char *buf = static_cast<char *>(bm->Bits ());
						if (buf != NULL) {
							char *data;
							ssize_t bytesRead;
							
							err = msg.FindData ("bits", B_RAW_TYPE, 0, &data, &bytesRead);
							if (err == B_OK) {
								memcpy (buf, data, rowBytes*height);
							}
						}
					}
				}
			}
		}
	}
	return bm;
}

bool HelloApplication::CreateImage (void *v) {
	HelloWindow		*aWindow;
	HelloView		*aView;
	BRect			aRect;
	status_t err;
	char *outName;
	char *info;
	int32 outVersion;
	DATAInfo *outInfo = NULL;
	int32 outNumInfo;
	int32 whichOne = -1;
	entry_ref *the_ref = (entry_ref *)(v);
	
	if (the_ref != NULL) {
		BFile inSource (the_ref, O_RDONLY);
		if (inSource.InitCheck () == B_OK) {
			BMallocIO outSource;
			err = DATAGetHandlers (inSource, NULL, outInfo, outNumInfo, 'TIFF');
			if (err == B_OK) {
				for (int i=0;i<outNumInfo;i++) {
					err = DATAGetHandlerInfo(outInfo[i].formatHandler, outName, info, outVersion);
					if (err == B_OK) {
						if (strcmp (outName, "TIFF EBitmap Handler") == 0) {
							whichOne = i;
							break;
						}
					}
				}
				if (whichOne != -1) {
					err = DATATranslate (inSource, &outInfo[whichOne], NULL, outSource, E_BITMAP, 0, "image/tiff");
				}
				else {
					err = DATATranslate (inSource, NULL, NULL, outSource, E_BITMAP, 0, "image/tiff");
				}
				// we can finally delete this
				delete [] outInfo;
			}
			else {
				err = DATATranslate (inSource, NULL, NULL, outSource, E_BITMAP, 0, "image/tiff");
			}
			if (err == B_OK) {
				// translation was OK
				BMessage msg;
				
				outSource.Seek (0, SEEK_SET); // This is important!
				if ((err = msg.Unflatten (&outSource)) == B_OK) {
					EBitmap *eBmp = dynamic_cast<EBitmap *>(instantiate_object (&msg));
					if (eBmp != NULL) {
						BBitmap *bm = eBmp->MakeABitmap ();
						if (bm != NULL) {
							aRect = bm->Bounds ();
							aRect.OffsetBy (100, 80);
							aWindow = new HelloWindow(aRect);
							
							// set up a rectangle and instantiate a new view
							// view rect should be same size as window rect but with left top at (0, 0)
							aRect = bm->Bounds ();
							aView = new HelloView(aRect, "HelloView", bm);
							
							// add view to window
							aWindow->AddChild(aView);
							
							// make window visible
							aWindow->Show();
										
							return false;
						}
					}
					else {
						debug_printf ("instatiate_object failed!", 0);
					}
				}
				else {
					debug_printf ("Unflatten didn't work: %x", err);
				}
			}
			else {
				debug_printf ("Translation didn't happen: %x", err);
			}
		}
	}
	// something went wrong
	be_app->PostMessage (B_QUIT_REQUESTED);
	return false;
}
#endif /* DEMO_5 */
