/*
 * handler.cpp
 *
 * Copyright 1997-1998 Geoffrey Clements
 *
 * License:
 * 		This code may be freely used in your programs.
 */

#define DEMO_1
//#define DEMO_2
//#define DEMO_3
//#define DEMO_4
//#define DEMO_5

#define debug_printf(s,n) { char buf[256]; sprintf(buf, s, n); debugger(buf); }

#ifndef _DATATYPES_H_
#include <Datatypes.h>
#endif
#ifndef _DATAHANDLER_H_
#include <DataHandler.h>
#endif
#ifndef _TIFF_
#include <tiff.h>
#endif
#ifndef _TIFFIO_
#include <tiffio.h>
#endif

#ifndef EBITMAP_H
#include "EBitmap.h"
#endif

#define	VERSION	"v1.0.0"
#define	EMAIL	"<clements@net1plus.com>"

char			handlerName[] = "TIFF Handler";
char			handlerInfo[] = "TIFF file format handler " VERSION " " EMAIL;
int32			handlerVersion = 100;

static	int	debug = 0;

/*
	uint32		type;						//	BeOS data type
	uint32		t_cls;						//	class of this format
	float		quality;					//	format quality 0.0-1.0
	float		capability;					//	handler capability 0.0-1.0
	char		MIME[B_MIME_TYPE_LENGTH+11];	//	MIME string
	char		formatName[B_MIME_TYPE_LENGTH+11];	//	only descriptive
*/
Format inputFormats[] = {
	{ 'TIFF', E_BITMAP, 0.6, 0.5, "image/tiff", "TIFF bitmap format" },
	{ E_BITMAP, E_BITMAP, 0.6, 0.6, "image/x-vnd.Be-EBitmap", "Extended Bitmap Format" },
	{ 0, 0, 0, 0, 0, 0 },
};

Format outputFormats[] = {
	{ 'TIFF', E_BITMAP, 0.6, 0.5, "image/tiff", "TIFF bitmap format" },
	{ E_BITMAP, E_BITMAP, 0.6, 0.6, "image/x-vnd.Be-EBitmap", "Extended Bitmap Format" },
	{ 0, 0, 0, 0, 0, 0 },
};

static status_t WriteBitmap (BPositionIO &inSource, BPositionIO &outDestination, BMessage *ioExtension);
static status_t WriteTIFFFile (BPositionIO &inSource, BPositionIO &outDestination, BMessage *ioExtension);
static status_t IdentifyTIFF (BPositionIO &inSource, DATAInfo &outInfo);
static status_t IdentifyBitmap (BPositionIO &inSource, DATAInfo &outInfo);
static status_t CopyLoop (BPositionIO &input, BPositionIO &output);

BMessage prefs;
// this contains all the prefernces
// entry_ref "Open Profile" profile to use on open
// entry_ref "Save Profile" profile to use when saving

//	Return DATA_NO_HANDLER if not handling this data.
//	Even if inputFormats exists, may be called for data without hints
//	Ff outType is not 0, must be able to export in wanted format

status_t Identify(BPositionIO &inSource, const Format *inFormat,
	BMessage *, DATAInfo &outInfo, uint32 outType) {
	status_t returnValue = DATA_NO_HANDLER;
	
	if ((outType == 0) || (outType == E_BITMAP) || (outType == 'TIFF')) {
		
		if (outType == E_BITMAP) { // input is a TIFF
			returnValue = IdentifyTIFF (inSource, outInfo);
		}
		else if (outType == 'TIFF') {
			returnValue = IdentifyBitmap (inSource, outInfo);
		}
		else if (outType == 0) {
			if (inFormat == NULL) {
				// assume its a input tiff file and wing it...
				returnValue = IdentifyTIFF (inSource, outInfo);
			}
			else {
				// want a default translation
				if (inFormat->type == 'TIFF') {
					returnValue = IdentifyTIFF (inSource, outInfo);
				}
				else if (inFormat->type == E_BITMAP) {
					returnValue = IdentifyBitmap (inSource, outInfo);
				}
				else {
					// assume its a input tiff file and wing it...
					returnValue = IdentifyTIFF (inSource, outInfo);
				}
			}
		}
	}
	
	return returnValue;
}

//	Return DATA_NO_HANDLER if not handling the output format
//	If outputFormats exists, will only be called for those formats

status_t Translate(BPositionIO &inSource, const DATAInfo &inInfo,
	BMessage *ioExtension, uint32 outFormat, BPositionIO &outDestination) {
	
	if (outFormat == inInfo.formatType) {
		return CopyLoop(inSource, outDestination);
	}
	else if ((outFormat == E_BITMAP) && (inInfo.formatType == 'TIFF')) {
		return WriteBitmap (inSource, outDestination, ioExtension);
	}
	else if ((outFormat == 'TIFF') && (inInfo.formatType == E_BITMAP)){
		return WriteTIFFFile (inSource, outDestination, ioExtension);
	}

	return DATA_NO_HANDLER;
}

#if !defined(DEMO_5)
status_t WriteBitmap (BPositionIO &inSource, BPositionIO &outDestination, BMessage *ioExtension) {
	TIFF *tif = TIFFOpen ((thandle_t)(&inSource));
	
	if (tif == NULL) {
		return DATA_NO_HANDLER;
	}
	
	BMessage eBmp;
	status_t err;
	ssize_t size;
	uint32 width;
	uint32 height;
	uint32 scanLineSize;
	uint32 rowBytes;
	tdata_t inBuf;
	rgb_color *outBuf;
	uint16 extraSamples = 0;
	
	//	set up header
	eBmp.what = E_BITMAP;
	TIFFGetField (tif, TIFFTAG_IMAGEWIDTH, &width);
	eBmp.AddInt32 ("width", width);
	TIFFGetField (tif, TIFFTAG_IMAGELENGTH, &height);
	eBmp.AddInt32 ("height", height);
	TIFFGetField (tif, TIFFTAG_EXTRASAMPLES, &extraSamples);
	
#if defined(DEMO_1)
	eBmp.AddInt32 ("storage", E_STORAGE_FOLLOWS);
	eBmp.AddPointer ("BPositionIO", &outDestination);
#elif defined(DEMO_2)
	eBmp.AddInt32 ("storage", E_STORAGE_POINTER);
	// pointer is added below
#elif defined(DEMO_3)
	eBmp.AddInt32 ("storage", E_STORAGE_AREA_ID);
	// area_id is added below
#elif defined(DEMO_4)
	eBmp.AddInt32 ("storage", E_STORAGE_SELF);
	// data is added below
#endif
	eBmp.AddInt32 ("photometric", B_RGB_32_BIT);
	rowBytes = width * sizeof (rgb_color);
	eBmp.AddInt32 ("row bytes", rowBytes);
	
#if defined(DEMO_1)
	//	we always output in 24 bit format, since palettes cannot be represented well in 8 bit mode
	err = eBmp.Flatten (&outDestination, &size);
	if (err < B_OK) return err;
	
	err = outDestination.SetSize(size + (sizeof(rgb_color)*width*height));
	if (err < B_OK) return err;
	
	scanLineSize = TIFFScanlineSize (tif);
	inBuf = _TIFFmalloc (scanLineSize);
	outBuf = static_cast<rgb_color *>(malloc (rowBytes));
	for (int32 row = 0; row < height; ++row) {
		TIFFReadScanline (tif, inBuf, row);
		uint8 *p = static_cast<uint8 *>(inBuf);
		for (int i=0;i<width;i++) {
			outBuf[i].blue = *p++;
			outBuf[i].green = *p++;
			outBuf[i].red = *p++;
			outBuf[i].alpha = 255;
			for (int j=0;j<extraSamples;j++) {
				p++;
			}
		}
		err = outDestination.Write(outBuf, rowBytes);
	}
	free (outBuf);
	_TIFFfree (inBuf);

#elif defined(DEMO_2)
	scanLineSize = TIFFScanlineSize (tif);
	inBuf = _TIFFmalloc (scanLineSize);

	char *ptr = static_cast<char *>(malloc (rowBytes*height));
	for (int32 row = 0; row < height; ++row) {
		outBuf = (rgb_color *)(ptr+(rowBytes*row)); // for some reason a static_cast doesn't work here
		TIFFReadScanline (tif, inBuf, row);
		uint8 *p = static_cast<uint8 *>(inBuf);
		for (int i=0;i<width;i++) {
			outBuf[i].blue = *p++;
			outBuf[i].green = *p++;
			outBuf[i].red = *p++;
			outBuf[i].alpha = 255;
			for (int j=0;j<extraSamples;j++) {
				p++;
			}
		}
	}
	_TIFFfree (inBuf);

	eBmp.AddPointer ("pointer", ptr);
	
	err = eBmp.Flatten (&outDestination, &size);
	if (err < B_OK) {
		free (ptr);
		return err;
	}
#elif defined(DEMO_3)
	int64 area_size = rowBytes*height;
	char *ptr;
	
	// round up to nearest page size
	area_size = (((area_size-1)%B_PAGE_SIZE)+1)*B_PAGE_SIZE;

	/* Make sure the size won't overflow a size_t spec. */ 
	if (area_size >= ((1LL<<32)-1) ) { 
		return B_NO_MEMORY; 
	}
	area_id area = create_area("image area", (void **)&ptr, B_ANY_ADDRESS, area_size, B_NO_LOCK, B_READ_AREA | B_WRITE_AREA);
	if (area < B_OK) {
		return area;
	}

	scanLineSize = TIFFScanlineSize (tif);
	inBuf = _TIFFmalloc (scanLineSize);
	if (inBuf == NULL) {
		delete_area (area);
		return B_NO_MEMORY;
	}
	for (int32 row = 0; row < height; ++row) {
		outBuf = (rgb_color *)(ptr+(rowBytes*row)); // for some reason a static_cast doesn't work here
		TIFFReadScanline (tif, inBuf, row);
		uint8 *p = static_cast<uint8 *>(inBuf);
		for (int i=0;i<width;i++) {
			outBuf[i].blue = *p++;
			outBuf[i].green = *p++;
			outBuf[i].red = *p++;
			outBuf[i].alpha = 255;
			for (int j=0;j<extraSamples;j++) {
				p++;
			}
		}
	}
	_TIFFfree (inBuf);

	eBmp.AddInt32 ("area_id", area);
	
	err = eBmp.Flatten (&outDestination, &size);
	if (err < B_OK) {
		free (ptr);
		return err;
	}
#elif defined(DEMO_4)
	scanLineSize = TIFFScanlineSize (tif);
	inBuf = _TIFFmalloc (scanLineSize);

	char *ptr = static_cast<char *>(malloc (rowBytes*height));
	for (int32 row = 0; row < height; ++row) {
		outBuf = (rgb_color *)(ptr+(rowBytes*row)); // for some reason a static_cast doesn't work here
		TIFFReadScanline (tif, inBuf, row);
		uint8 *p = static_cast<uint8 *>(inBuf);
		for (int i=0;i<width;i++) {
			outBuf[i].blue = *p++;
			outBuf[i].green = *p++;
			outBuf[i].red = *p++;
			outBuf[i].alpha = 255;
			for (int j=0;j<extraSamples;j++) {
				p++;
			}
		}
	}
	_TIFFfree (inBuf);
	
	eBmp.AddData ("bits", B_RAW_TYPE, ptr, rowBytes*height);
	free (ptr);

	err = eBmp.Flatten (&outDestination, &size);
	if (err < B_OK) {
		return err;
	}
#endif

	TIFFClose (tif);
	return B_OK;
}
#else
// demo 5
class EBitmap : public BArchivable {
	TIFF *tif;
	BMessage eBmp;
	
public:
	EBitmap ();
	~EBitmap ();
	
	status_t Initialize (BPositionIO &inSource);
	
	virtual	status_t Archive (BMessage *into, bool deep = true) const;
};

EBitmap::EBitmap () : tif (NULL) {
	
	eBmp.MakeEmpty ();
}

EBitmap::~EBitmap () {
	
	if (tif != NULL) {
		TIFFClose (tif);
	}
}

status_t EBitmap::Archive (BMessage *into, bool deep) const {
	status_t err;
	
	// copy all our stuff into this message
	*into = eBmp;
	// add our class name to the array
	if ((err = into->AddString ("class", "EBitmap")) == B_OK) {
		// call the inherited version
		err = BArchivable::Archive (into, deep);
	}
	return err;	
}

status_t EBitmap::Initialize (BPositionIO &inSource) {

	tif = TIFFOpen ((thandle_t)(&inSource));
	
	if (tif == NULL) {
		return DATA_NO_HANDLER;
	}
	
	uint32 width;
	uint32 height;
	uint32 scanLineSize;
	uint32 rowBytes;
	tdata_t inBuf;
	rgb_color *outBuf;
	uint16 extraSamples = 0;
	
	//	set up header
	eBmp.what = E_BITMAP;
	TIFFGetField (tif, TIFFTAG_IMAGEWIDTH, &width);
	eBmp.AddInt32 ("width", width);
	TIFFGetField (tif, TIFFTAG_IMAGELENGTH, &height);
	eBmp.AddInt32 ("height", height);
	TIFFGetField (tif, TIFFTAG_EXTRASAMPLES, &extraSamples);

	eBmp.AddInt32 ("storage", E_STORAGE_SELF);
	// data is added below

	eBmp.AddInt32 ("photometric", B_RGB_32_BIT);
	rowBytes = width * sizeof (rgb_color);
	eBmp.AddInt32 ("row bytes", rowBytes);
	
	scanLineSize = TIFFScanlineSize (tif);
	inBuf = _TIFFmalloc (scanLineSize);

	char *ptr = static_cast<char *>(malloc (rowBytes*height));
	for (int32 row = 0; row < height; ++row) {
		outBuf = (rgb_color *)(ptr+(rowBytes*row)); // for some reason a static_cast doesn't work here
		TIFFReadScanline (tif, inBuf, row);
		uint8 *p = static_cast<uint8 *>(inBuf);
		for (int i=0;i<width;i++) {
			outBuf[i].blue = *p++;
			outBuf[i].green = *p++;
			outBuf[i].red = *p++;
			outBuf[i].alpha = 255;
			for (int j=0;j<extraSamples;j++) {
				p++;
			}
		}
	}
	_TIFFfree (inBuf);
	
	eBmp.AddData ("bits", B_RAW_TYPE, ptr, rowBytes*height);
	free (ptr);
		
	return B_OK;
}

status_t WriteBitmap (BPositionIO &inSource, BPositionIO &outDestination, BMessage *ioExtension) {
	EBitmap eBmp;
	status_t err;
	ssize_t size;
	
	if ((err = eBmp.Initialize (inSource)) == B_OK) {
		BMessage archive;
	
		err = eBmp.Archive (&archive, true);
		if (err == B_OK) {
			err = archive.Flatten (&outDestination, &size);
		}
	}
	return err;
}

#endif /* DEMO_5 */

status_t WriteTIFFFile (BPositionIO &inSource, BPositionIO &outDestination, BMessage *ioExtension) {
	TIFF *tif = TIFFOpen ((thandle_t)(&outDestination));
	
	if (tif == NULL) {
		return DATA_NO_HANDLER;
	}
	
	BMessage eBmp;
	status_t err;
	err = eBmp.Unflatten (&inSource);
	if (err < B_OK) return err;
	int32 eBmpPhotometric;
	err = eBmp.FindInt32 ("photometric", &eBmpPhotometric);
	
	if (eBmpPhotometric != B_RGB_32_BIT) {
		TIFFClose (tif);
		return DATA_NO_HANDLER;
	}

	int32 width;
	int32 height;
	if (eBmp.FindInt32 ("width", &width) == B_OK) TIFFSetField (tif, TIFFTAG_IMAGEWIDTH, width);
	else return DATA_NO_HANDLER;
	if (eBmp.FindInt32 ("height", &height) == B_OK) TIFFSetField (tif, TIFFTAG_IMAGELENGTH, height);
	else return DATA_NO_HANDLER;
	
	int32 storage;
	if ((eBmp.FindInt32 ("storage", &storage) != B_OK) || (storage != E_STORAGE_FOLLOWS))
		return DATA_NO_HANDLER;

	uint16 bitsPerSample = 8;
	uint16 samplesPerPixel = 3;
	uint16 extraSamples = 0;
	uint16 photometric = PHOTOMETRIC_RGB;
	int32 planarConfig = PLANARCONFIG_CONTIG;
	
	TIFFSetField (tif, TIFFTAG_BITSPERSAMPLE, bitsPerSample);
	TIFFSetField (tif, TIFFTAG_SAMPLESPERPIXEL, samplesPerPixel);
	TIFFSetField (tif, TIFFTAG_EXTRASAMPLES, extraSamples);
	TIFFSetField (tif, TIFFTAG_PHOTOMETRIC, photometric);
	TIFFSetField (tif, TIFFTAG_PLANARCONFIG, planarConfig); // PLANARCONFIG_CONTIG = 1
	// If this isn't set the TIFF image written will not be valid
	// about a megabyte per strip
	int64 rowsPerStrip = ((int64)samplesPerPixel*(int64)width*(int64)height)/(int64)1000000;
	int32 rps = static_cast<int32>(rowsPerStrip);
	if (rps < 1) rps = 1;
	TIFFSetField (tif, TIFFTAG_ROWSPERSTRIP, rps);
	
	// need to write the TIFF data here.
	uint32 scanlineLength = 4*width;
	uint8 *outBuf = static_cast<uint8 *>(malloc (scanlineLength));
	for (int i=0; i < height; i++) {
		inSource.Read (outBuf, scanlineLength);
		uint8 *p = outBuf;
		uint8 *p2 = outBuf;
		// TIFF images expect RGB we seem to be getting BGRA
		for (int j=0;j<width;j++) {
			*p++ = *(p2+2);
			*p++ = *(p2+1);
			*p++ = *p2;
			p2+=4;
		}
		TIFFWriteScanline (tif, outBuf, i);
	}
	free (outBuf);
	TIFFFlush (tif);
	TIFFClose (tif);
	return B_OK;
}

//	The view will get resized to what the parent thinks is 
//	reasonable. However, it will still receive MouseDowns etc.
//	Your view should change settings in the translator immediately, 
//	taking care not to change parameters for a translation that is 
//	currently running. Typically, you'll have a global struct for 
//	settings that is atomically copied into the translator function 
//	as a local when translation starts.
//	Store your settings wherever you feel like it.

#include "ConfigView.h"

status_t MakeConfig (BMessage *ioExtension, BView * &outView, BRect &outExtent) {

	// if ioExtension is not null make a copy of it and use it in attached to window
	outView = new ConfigView ("main", NULL);
	outExtent = outView->Bounds ();
	
	return B_OK;
}


	//	Copy your current settings to a BMessage that may be passed 
	//	to DATATranslate at some later time when the user wants to 
	//	use whatever settings you're using right now.

status_t GetConfigMessage (BMessage *ioExtension) {
	
	return B_OK;
}

status_t IdentifyTIFF (BPositionIO &inSource, DATAInfo &outInfo) {
	TIFF *tif = TIFFOpen ((thandle_t)(&inSource));
	
	if (tif == NULL) {
		return DATA_NO_HANDLER;
	}
	// for now all we accept is plain RGB images
	uint16 bitsPerSample;
	uint16 photometric;
	uint16 planarConfig;
	
	TIFFGetField (tif, TIFFTAG_PLANARCONFIG, &planarConfig); // PLANARCONFIG_CONTIG = 1
	if (planarConfig != PLANARCONFIG_CONTIG) return DATA_NO_HANDLER; 
	TIFFGetField (tif, TIFFTAG_BITSPERSAMPLE, &bitsPerSample);
	if (bitsPerSample != 8) return DATA_NO_HANDLER;
	if (TIFFGetField (tif, TIFFTAG_PHOTOMETRIC, &photometric) && (photometric != PHOTOMETRIC_RGB)) return DATA_NO_HANDLER;
	
	outInfo.formatType = inputFormats[0].type;
	outInfo.formatHandler = 0;  // handled by libdatatypes.so
	outInfo.formatGroup = inputFormats[0].t_cls;
	outInfo.formatQuality = inputFormats[0].quality;
	outInfo.formatCapability = inputFormats[0].capability;
	strcpy (outInfo.MIMEName, inputFormats[0].MIME);
	strcpy (outInfo.formatName, inputFormats[0].formatName);
	
	TIFFClose (tif);
	
	return B_OK;
}

status_t IdentifyBitmap (BPositionIO &inSource, DATAInfo &outInfo) {
	BMessage eBmp;
	
	status_t err = eBmp.Unflatten (&inSource);
	if ((err != B_OK) && (eBmp.what != E_BITMAP)) {
		return DATA_NO_HANDLER;
	}
	
	// asume the rest of the message is valid. We should probably check
	// to make sure required EBitmap entries are present.
	// width, height, photometric, storage What the heck we will
	int32 value;
	
	if ((B_OK != eBmp.FindInt32 ("width", &value)) ||
		(B_OK != eBmp.FindInt32 ("height", &value)) ||
		(B_OK != eBmp.FindInt32 ("photometric", &value)) ||
		(B_OK != eBmp.FindInt32 ("storage", &value))) {
		// we should make sure the values are valid
		return DATA_NO_HANDLER;
	}
	
	outInfo.formatType = outputFormats[0].type;
	outInfo.formatHandler = 0;  // handled by libdatatypes.so
	outInfo.formatGroup = outputFormats[0].t_cls;
	outInfo.formatQuality = outputFormats[0].quality;
	outInfo.formatCapability = outputFormats[0].capability;
	strcpy (outInfo.MIMEName, outputFormats[0].MIME);
	strcpy (outInfo.formatName, outputFormats[0].formatName);
	
	return B_OK;
}

static status_t CopyLoop (BPositionIO &input, BPositionIO &output) {
	size_t block = 65536L;
	void * buffer = malloc(block);
	char temp[1024];
	
	if (!buffer) {
		buffer = temp;
		block = 1024;
	}
	bool done = false;
	status_t err = B_OK;
	while (!done) {
		ssize_t to_read = block;
		err = input.Read(buffer, to_read);
		if (err < B_OK) break;
		done = (to_read < block);
		if (debug) printf("Write %d\n", to_read);
		err = output.Write(buffer, to_read);
		if (err < B_OK) break;
	}
	if (buffer != temp)
		free(buffer);
	return err > 0 ? B_OK : err;
}
